@@ -21,23 +21,17 @@
 #include <assert.h>
 #include <klocale.h>
 #include <kpushbutton.h>
+#include <qalgorithms.h>
 
 #include "src/logic/elapsedtime.h"
 #include "src/logic/levelloader.h"
 
+static QString diffString(const int difficulty);
+
 class LevelTableModel : public QAbstractTableModel
 {
 public:
-    LevelTableModel(const QList<QSharedPointer<Level> > &levels, QObject * parent = 0);
-
-protected:
-    int rowCount(const QModelIndex &parent) const;
-    int columnCount(const QModelIndex &parent) const;
-    QVariant data(const QModelIndex &index, int role) const;
-    QVariant headerData(int section, Qt::Orientation orientation, int role) const;
-
-private:
-    const QList<QSharedPointer<Level> > &m_levels;
+    LevelTableModel(QList<QSharedPointer<Level> > &levels, QObject * parent = 0);
 
     enum Columns {
         Name,
@@ -47,9 +41,18 @@ private:
         Solved,
         ColumnCount /* not a real column */
     };
+
+    int rowCount(const QModelIndex &parent = QModelIndex()) const;
+    int columnCount(const QModelIndex &parent = QModelIndex()) const;
+    QVariant data(const QModelIndex &index, int role) const;
+    QVariant headerData(int section, Qt::Orientation orientation, int role) const;
+    void sort(int column, Qt::SortOrder order = Qt::AscendingOrder);
+
+private:
+    QList<QSharedPointer<Level> > &m_levels;
 };
 
-LevelTableModel::LevelTableModel(const QList<QSharedPointer<Level> > &levels, QObject *parent) :
+LevelTableModel::LevelTableModel(QList<QSharedPointer<Level> > &levels, QObject *parent) :
     QAbstractTableModel(parent), m_levels(levels)
 {
 
@@ -78,9 +81,9 @@ QVariant LevelTableModel::data(const QModelIndex &index, int role) const {
     switch (index.column()) {
     case Name: return level->visibleName();
     case LevelSet: return level->levelset();
-    case Difficulty: return level->difficulty();
+    case Difficulty: return diffString(level->difficulty());
     case Size: return QString("%1x%2").arg(level->width()).arg(level->height());
-    case Solved: return "TODO";
+    case Solved: return (level->solved()) ? Time(level->solvedTime()).toString() : "-";
     default: assert(0);
     }
 
@@ -101,28 +104,101 @@ QVariant LevelTableModel::headerData(int section, Qt::Orientation orientation, i
     return QAbstractTableModel::headerData(section, orientation, role);
 }
 
+typedef bool (*LevelComparator)(const QSharedPointer<Level> &,
+                                const QSharedPointer<Level> &);
+
+static bool levelLessThan(const QSharedPointer<Level> &lhs,
+                          const QSharedPointer<Level> &rhs) {
+    return lhs->name() < rhs->name();
+}
+
+static bool levelGreaterThan(const QSharedPointer<Level> &lhs,
+                             const QSharedPointer<Level> &rhs) {
+    return levelLessThan(rhs, lhs);
+}
+
+static bool diffLessThan(const QSharedPointer<Level> &lhs,
+                         const QSharedPointer<Level> &rhs) {
+    return lhs->difficulty() < rhs->difficulty();
+}
+
+static bool diffGreaterThan(const QSharedPointer<Level> &lhs,
+                            const QSharedPointer<Level> &rhs) {
+    return diffLessThan(rhs, lhs);
+}
+
+static bool solvedLessThan(const QSharedPointer<Level> &lhs,
+                           const QSharedPointer<Level> &rhs) {
+    if (!lhs->solved() && !rhs->solved()) {
+        return false;
+    } else if (!lhs->solved() || !rhs->solved()) {
+        return lhs->solved();
+    } else {
+        return (lhs->solvedTime() < rhs->solvedTime());
+    }
+}
+
+static bool solvedGreaterThan(const QSharedPointer<Level> &lhs,
+                              const QSharedPointer<Level> &rhs) {
+    return solvedLessThan(rhs, lhs);
+}
+
+void LevelTableModel::sort(int column, Qt::SortOrder order) {
+    LevelComparator cmp = NULL;
+    switch (column) {
+    case Name: cmp = (order == Qt::AscendingOrder) ? levelLessThan : levelGreaterThan; break;
+    case Difficulty: cmp = (order == Qt::AscendingOrder) ? diffLessThan : diffGreaterThan; break;
+    case Solved: cmp = (order == Qt::AscendingOrder) ? solvedLessThan : solvedGreaterThan; break;
+    default: assert(0);
+    }
+
+    qStableSort(m_levels.begin(), m_levels.end(), cmp);
+    emit dataChanged(index(0, 0), index(rowCount() - 1, columnCount() -1));
+}
+
 SelectBoardWindow::SelectBoardWindow(QWidget *parent)
     : KDialog(parent)
 {
     setModal(true);
     setCaption(i18n("Level Selection"));
-    setMinimumSize(600, 350);
 
     ui = new Ui::LevelSelectUi;
     ui->setupUi(mainWidget());
 
     m_levels = LevelLoader::load();
     m_model = QSharedPointer<LevelTableModel>(new LevelTableModel(m_levels));
-    ui->listView->setModel(m_model.data());
+
+    ui->tableView->setUpdatesEnabled(false);
+    ui->tableView->setModel(m_model.data());
+
+    /* Hide all columns except Level, Difficulty and Solved. */
+
+    for (int i = 0; i < m_model->columnCount(); i++) {
+        ui->tableView->hideColumn(i);
+    }
+    ui->tableView->showColumn(LevelTableModel::Name);
+    ui->tableView->showColumn(LevelTableModel::Difficulty);
+    ui->tableView->showColumn(LevelTableModel::Solved);
+
+    ui->tableView->sortByColumn(LevelTableModel::Name, Qt::AscendingOrder);
+    ui->tableView->sortByColumn(LevelTableModel::Solved, Qt::DescendingOrder);
+    ui->tableView->sortByColumn(LevelTableModel::Difficulty, Qt::AscendingOrder);
+
+    ui->tableView->resizeColumnsToContents();
+    ui->tableView->resizeRowsToContents();
 
     if (m_levels.empty()) {
         button(KDialog::Ok)->setEnabled(false);
     } else {
-        QModelIndex index = m_model->index(0, 0);
-        ui->listView->selectionModel()->select(index, QItemSelectionModel::Select);
-        connect(ui->listView->selectionModel(), SIGNAL(currentRowChanged(QModelIndex,QModelIndex)), this, SLOT(selectedLevelChanged(QModelIndex,QModelIndex)));
+        resetSelection();
+        connect(ui->tableView->selectionModel(), SIGNAL(currentRowChanged(QModelIndex,QModelIndex)),
+                this, SLOT(selectedLevelChanged(QModelIndex,QModelIndex)));
+        connect(m_model.data(), SIGNAL(dataChanged(QModelIndex,QModelIndex)),
+                this, SLOT(levelDataChanged(QModelIndex,QModelIndex)));
         updateDetails(m_levels[0]);
     }
+
+    ui->tableView->setUpdatesEnabled(true);
 }
 
 void SelectBoardWindow::showEvent(QShowEvent *event) {
@@ -144,7 +220,15 @@ void SelectBoardWindow::selectedLevelChanged(const QModelIndex &current, const Q
     updateDetails(m_levels[current.row()]);
 }
 
-QString SelectBoardWindow::diffString(int difficulty) const {
+void SelectBoardWindow::levelDataChanged(const QModelIndex &topLeft, const
+                                         QModelIndex &bottomRight) {
+    Q_UNUSED(topLeft);
+    Q_UNUSED(bottomRight);
+    resetSelection();
+    updateDetails(selectedBoard());
+}
+
+static QString diffString(const int difficulty) {
     if (difficulty < 0) {
         return diffString(0);
     } else if (difficulty > 7) {
@@ -182,7 +266,14 @@ void SelectBoardWindow::updateDetails(QSharedPointer<Level> level) {
     }
 }
 
+void SelectBoardWindow::resetSelection() {
+    QModelIndex index = m_model->index(0, 0);
+    QItemSelectionModel::SelectionFlags flags =
+            QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Rows;
+    ui->tableView->selectionModel()->select(index, flags);
+}
+
 QSharedPointer<Level> SelectBoardWindow::selectedBoard() const {
-    int index = ui->listView->selectionModel()->selectedIndexes().at(0).row();
+    int index = ui->tableView->selectionModel()->selectedIndexes().at(0).row();
     return m_levels[index];
 }
